
package org.biospice.moma;
import com.sri.oaa2.icl.*;
import com.sri.oaa2.lib.*;
import java.net.*;
import java.io.*;
import java.util.*;

/**
 * A service that performs MOMA calculations by calling an appropriate perl script.
 * 
 * @author <a href="mailto:pkharch@fas.harvard.edu">Peter Kharchenko</a>
 * @version 1.0
 */
public class MomaService implements OAAEventListener {
    public final static String mfName="moma"; // name of the main method call
    public final static String nfName="moma_gene_names"; // name of the utility method for obtaining gene name sequence
    public final static String mgfName="moma_mutable_genes"; // name of the utility method for obtaining gene name sequence

    public final static String tempDirPrefix="session";


    public final static String propertiesFileName="moma.properties";

    private static long counter=0; 
    Properties props=new Properties();

    Map geneNamesMap=new Hashtable(); // a table for looking up gene number given its name
    IclInt growthFluxNumber; // a flux (gene) number corresponding to the flux
    IclList geneNamesList=new IclList();
    IclList mutableGeneNumbers=new IclList();
    IclList mutableGenePrettyNames=new IclList();
    String perlFileCanonicalPath=null;

    String inputFileName=null;

    public MomaService() throws Exception {
        // load properties file
        try {
            props.load(new FileInputStream(propertiesFileName));
        } catch (FileNotFoundException fnfe) {
            System.err.println("Unable to open properties file \""+propertiesFileName+"\"");
            throw fnfe;
        } catch (IOException ioe) {
            System.err.println("Unable to read properties file \""+propertiesFileName+"\"");
            throw ioe;
        }

        // look up gene names file
        String geneNamesFileName=props.getProperty("GENE_NAMES_FILE");
        inputFileName=props.getProperty("INPUT_FILE_NAME");

        // obtain canonical path for the 
        File pf=new File(props.getProperty("MOMA_PERL_FILE")).getCanonicalFile();
        perlFileCanonicalPath=pf.getCanonicalPath();
        
        // read gene names list
        try {
            BufferedReader br=new BufferedReader(new FileReader(geneNamesFileName));
            // first thing in the file is the growth flux number
            growthFluxNumber=new IclInt(Integer.parseInt(br.readLine()));
        
            // read in all of the gene names
            int gene_number=1;

            String line=br.readLine();
            while(line!=null) {
                String geneName=line.trim();
                geneNamesMap.put(geneName,new Integer(gene_number));
                geneNamesList.add(new IclStr(geneName));

                line=br.readLine();
            }
        } catch (FileNotFoundException fnfe2) {
            System.err.println("Unable to open gene names file \""+geneNamesFileName+"\"");
            throw fnfe2;
        } catch (IOException ioe2) {
            System.err.println("Unable to read gene names file \""+geneNamesFileName+"\"");
            throw ioe2;
        }

        // read mutable gene info
        String mutableGenesFileName=props.getProperty("MUTABLE_GENES_FILE");
        try {
            BufferedReader br=new BufferedReader(new FileReader(mutableGenesFileName));
            String line=br.readLine();
            while(line!=null) {
                StringTokenizer st=new StringTokenizer(line);
                IclInt geneNumber=new IclInt(Integer.parseInt(st.nextToken()));
                mutableGeneNumbers.add(geneNumber);
                IclStr prettyGeneName=new IclStr(st.nextToken());
                mutableGenePrettyNames.add(prettyGeneName);
                line=br.readLine();
            }
        } catch (FileNotFoundException fnfe3) {
            System.err.println("Unable to open mutable genes file \""+mutableGenesFileName+"\"");
            throw fnfe3;
        } catch (IOException ioe3) {
            System.err.println("Unable to read mutable genes file \""+mutableGenesFileName+"\"");
            throw ioe3;
        }
    }

    public boolean doOAAEvent(IclTerm goal, IclList param, IclList answers) {
        System.out.println("received a request for a goal=\""+goal.toString()+"\"");
        
        if(goal.isStruct() && ((IclStruct)goal).getFunctor().equals(mfName)) {
            // read the isoenzyme setting
            IclTerm argument0=goal.getTerm(0);
            boolean mutateIsozymes=true;
            if(argument0.isInt()) {
                int value=((IclInt)argument0).toInt();
                if(value==0) {
                    mutateIsozymes=false;
                }
            } else {
                System.out.println("first argument must be an integer (0 or 1) !");
                return false;
            }
            
            // get a list of uptake rates

            IclTerm argument1=goal.getTerm(1);
            if(!(argument1.isList() && argument1.size()==3)) {
                System.out.println("second argument must be a list of three floating point values!");
                return false;
            }

            // get list of mutated genes
            IclTerm argument2=goal.getTerm(2);
            if(argument2.isList()) {

                // obtain counter value
                long counterValue=getAndIncrementCounter();
                // create new temp directory
                File tempDirectory=new File(tempDirPrefix+"_"+Long.toString(counterValue));
                tempDirectory.mkdir();
                try {
                    PrintWriter input_pw=new PrintWriter(new FileWriter(new File(tempDirectory,inputFileName)));
                    // print isoenzyme selection
                    if(mutateIsozymes) {
                        input_pw.println("1");
                    } else {
                        input_pw.println("0");
                    }
                    // print uptake rates
                    for(int i=0;i<3;i++) {
                        IclTerm t=argument1.getTerm(i);
                        // should probably check for number format here
                        input_pw.println(t.toString());
                    }

                    for(int i=0;i<argument2.size();i++) {
                        IclTerm t=argument2.getTerm(i);
                        if(t.isInt()) {
                            input_pw.println(Integer.toString(((IclInt)t).toInt()));
                        } else {
                            System.out.println("mutant gene list contains something other then integer at position "+Integer.toString(i)+"!");
                            input_pw.close();
                            deleteDir(tempDirectory);
                            return false;
                        }
                    }
                    input_pw.close();


                    // now that the input file has been created, we can call the perl script
                    Process p=Runtime.getRuntime().exec(perlFileCanonicalPath,null,tempDirectory.getCanonicalFile());
                    // wait for the script to finish
                    p.waitFor();
                
                    // read the wt results
                    IclList wtFluxes=readResults(new File(tempDirectory,props.getProperty("WT_FLUX_OUTPUT_NAME")));
                    IclList lpFluxes=readResults(new File(tempDirectory,props.getProperty("LP_FLUX_OUTPUT_NAME")));
                    IclList qpFluxes=readResults(new File(tempDirectory,props.getProperty("QP_FLUX_OUTPUT_NAME")));
                    answers.add(new IclStruct(mfName,argument0,argument1,argument2,wtFluxes,lpFluxes,qpFluxes));

                    deleteDir(tempDirectory);
                    return true;
                } catch (Exception e) {
                    e.printStackTrace();
                } finally {
                    deleteDir(tempDirectory);
                }
            } else {
                System.out.println("second argument must be an list of integers !");
                return false;
            }                

        } else if(goal.isStruct() && ((IclStruct)goal).getFunctor().equals(nfName)) {
            answers.add(new IclStruct(nfName,geneNamesList,growthFluxNumber));
        } else if(goal.isStruct() && ((IclStruct)goal).getFunctor().equals(mgfName)) {
            answers.add(new IclStruct(mgfName,mutableGeneNumbers,mutableGenePrettyNames));
        } else {
            System.out.println("unknown goal!");
        }
        return false;
    }

    /**
     * A utility method to read IclList of IclFloat from a text file, separated by "\n"
     *
     * @param f a <code>File</code> value
     * @return a <code>IclList</code> value
     */
    private IclList readResults(File f) throws FileNotFoundException, IOException{
        IclList v=new IclList();
        BufferedReader br=new BufferedReader(new FileReader(f));
        String line=br.readLine();
        while(line!=null) {
            v.add(new IclFloat(Double.parseDouble(line)));
            line=br.readLine();
        }
        br.close();
        return v;
    }

    private synchronized long getAndIncrementCounter() {
        counter++;
        return counter-1;
    }
    
    private static boolean deleteDir(File dir) {
/* 	if (dir.isDirectory()) {
            String[] children = dir.list();
            for (int i=0; i<children.length; i++) {
                boolean success = deleteDir(new File(dir, children[i]));
                if (!success) {
                    return false;
                }
            }
        }
    
        // The directory is now empty so delete it
        return dir.delete();  */
	return true;
    }

}
